import warnings
from importlib.util import find_spec
from os import environ
from pathlib import Path

from deepmerge import Merger

from appcfg.exceptions import (
    ConfigFileNotFoundError,
    InvalidModuleNameError,
    YamlExtraRequiredError,
)


def get_module_path(module_name: str) -> Path:
    """
    Given a module's name, returns the path of the source file associated with it
    """
    spec = find_spec(module_name)
    if spec is None:
        raise InvalidModuleNameError(module_name=module_name)
    return Path(spec.origin)


def get_root_path(module_path: Path, module_name: str) -> Path:
    """
    Given a module's path and name, returns the root directory of the project the module
    belongs to
    """
    depth = len(module_name.split(".")) - 1

    if depth == 0 or module_path.name == "__init__.py":
        # module_name refers to a standalone module / script (depth=0), or a package
        return module_path.parents[depth]

    # module_name refers to a module in a package
    return module_path.parents[depth - 1]


def get_config_dir(module_name) -> Path:
    """
    Given a module's name, returns the path of the corresponding config directory or
    raises a `FileNotFoundError` if the directory does not exist.
    """
    path = get_root_path(get_module_path(module_name), module_name) / "config"
    return path.resolve(strict=True)


def load_config_file(config_dir: Path, name: str, strict=False):
    """
    Parses a json or yaml file with the base name `name` from `config_dir`, if a
    corresponding config file exists there. Otherwise, `None` is returned if `strict` is
    `False`, and a `FileNotFoundError` is raised if `strict` is `True`.
    """
    base_path = config_dir / name

    for suffix in ["json", "yml", "yaml"]:
        path = base_path.with_suffix("." + suffix)
        if path.is_file():
            with path.open() as f:
                if suffix == "json":
                    import json

                    return json.load(f)
                else:
                    try:
                        import yaml
                    except ModuleNotFoundError:
                        raise YamlExtraRequiredError()

                    return yaml.safe_load(f)

    if strict:
        raise ConfigFileNotFoundError(base_path.as_posix() + ".{json,yml,yaml}")


def get_environment():
    """
    If an `ENV`, `PY_ENV`, or `ENVIRONMENT` environment variable is set, return the
    value of it. Otherwise, return "default".

    Special cases:
        * "dev" and "develop" are mapped to "development"
    """
    value = None
    for key in ["ENV", "PY_ENV", "ENVIRONMENT"]:
        if key in environ:
            value = environ[key]
            break
    if value is None or value == "":
        return "default"

    if value in {"dev", "develop"}:
        return "development"

    return value


def merge_configs(base: dict, override: dict):
    """
    Recursively merge `override` into `base`.
    """
    return Merger(
        [(list, ["override"]), (dict, ["merge"])], ["override"], ["override"],
    ).merge(base, override)


def validate_env_vars_template(template: dict):
    """
    Recursively check if the given dict contains only non-space-containing string
    values. If any other value is found, a warning is issued and the corresponding key
    is removed.
    """

    def validate(tpl: dict, path: str = ""):
        if path != "":
            path += "."
        for key, value in list(tpl.items()):
            key_path = path + key
            if isinstance(value, dict):
                validate(value, key_path)
                continue

            if not isinstance(value, str):
                warnings.warn(
                    (
                        "The env-vars config file contains a value of type {type} at {key}. "
                        "Only string values can be used to specify environment variables. "
                        "Ignoring {key}."
                    ).format(type=type(value), key=key_path)
                )
                tpl.pop(key)
                continue

            if " " in value:
                warnings.warn(
                    (
                        "Environment variables may not contain spaces. "
                        'Ignoring "{value}" at {key} in env-vars config file.'
                    ).format(value=value, key=key_path)
                )
                tpl.pop(key)
                continue

    validate(template)


def compile_env_vars_template(template: dict):
    """
    Given a (possibly recursive) dict that contains environment variable names, replace
    each name by the corresponding environmen variable's value, if set, or pop the key
    from the dict otherwise. Finally, recursively prune empty values from the dict.
    """

    def compile(tpl: dict):
        for key, value in list(tpl.items()):
            if isinstance(value, dict):
                compile(value)
                continue

            # values has to be a string at this point
            if value in environ:
                tpl[key] = environ[value]
            else:
                tpl.pop(key)

    compile(template)

    def prune(tpl: dict):
        """
        Recursively remove empty dicts and None values from a dict.
        """
        for key, value in list(tpl.items()):
            if isinstance(value, dict):
                if len(value) > 0:
                    prune(value)
                if len(value) == 0:
                    tpl.pop(key)

    prune(template)



from appcfg.util import (
    compile_env_vars_template,
    get_config_dir,
    get_environment,
    load_config_file,
    merge_configs,
    validate_env_vars_template,
)
from typing import Dict

# Map config directories to the corresponding config objects
_config_cache: Dict[str, dict] = {}


def _get_config(config_dir):
    """
    Given a config directory, assemble and return the corresponding configuration dict
    """
    config = load_config_file(config_dir, "default", strict=True)

    env = get_environment()
    if env != "default":
        override = load_config_file(config_dir, env)
        if override is not None:
            merge_configs(config, override)

    env_vars_override = load_config_file(config_dir, "env-vars")
    if env_vars_override is not None:
        validate_env_vars_template(env_vars_override)
        compile_env_vars_template(env_vars_override)
        merge_configs(config, env_vars_override)

    return config


# Public caching wrapper around _get_config
def get_config(module_name: str, cached=True):
    """
    Returns a dict that contains the content of `default.json` or `default.yml` in the `config` directory within the root module's directory, inferred from `module_name`.

    If an `ENV`, `PY_ENV`, or `ENVIRONMENT` variable is set (listed in the order of
    precedence), and a config file with a base name corresponding to that variable's
    value is found, the contents of that config file are merged into the default
    configuration. Additionally, the environment variables specified in `env-vars.json`
    or `env-vars.yml` override any other configuration when they are set.

    If none of `ENV`, `PY_ENV`, or `ENVIRONMENT` is set, only the `default` config file
    will be loaded and optionally be overridden by custom environment variables as
    specified in the `env-vars` config file. The `ENV` values "dev" and "develop" map to
    the `development.json` or `development.yml` config file.

    **Arguments**:

    - `module_name` (`str`): The name of the module (or any of its submodules) for which
      the configuration should be loaded for. `__name__` can be used when `get_config()`
      is called from the Python project that contains the `config` directory. Note that
      `appcfg` requires the `config` directory to be a direct child of the top-level
      package directory, or, if not a package, the directory that contains the specified
      module.

    - `cached` (`bool`): If `True` (the default), the configuration for each `config`
      directory will only be loaded once and the same dict object will be returned by
      every subsequent call to `get_config()`.
    """
    config_dir = get_config_dir(module_name)
    if not cached:
        return _get_config(config_dir)
    else:
        if config_dir in _config_cache:
            return _config_cache[config_dir]
        else:
            config = _get_config(config_dir)
            _config_cache[config_dir] = config
            return config



import os
from errno import ENOENT


class ConfigFileNotFoundError(FileNotFoundError):
    def __init__(self, path: str):
        super(ConfigFileNotFoundError, self).__init__(ENOENT, os.strerror(ENOENT), path)


class InvalidModuleNameError(ValueError):
    def __init__(self, module_name: str):
        super(InvalidModuleNameError, self).__init__(
            "The module name {} could not be resolved".format(module_name)
        )


class YamlExtraRequiredError(Exception):
    def __init__(self):
        super(YamlExtraRequiredError, self).__init__(
            'In order to parse YAML files, appcfg has to be installed with the ["yaml"] extra.'
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['appcfg']

package_data = \
{'': ['*']}

install_requires = \
['deepmerge>=0.1.0,<0.2.0']

extras_require = \
{'yaml': ['pyyaml>=5.3.1,<6.0.0']}

setup_kwargs = {
    'name': 'appcfg',
    'version': '1.0.0',
    'description': 'Flexible hierarchic Python application configuration',
    'long_description': '# appcfg\n\n[![PyPI version fury.io](https://badge.fury.io/py/appcfg.svg)](https://pypi.python.org/pypi/appcfg/)\n[![Build Status](https://travis-ci.com/bjoluc/appcfg.svg?branch=master)](https://travis-ci.com/bjoluc/appcfg)\n[![codecov](https://codecov.io/gh/bjoluc/appcfg/branch/master/graph/badge.svg)](https://codecov.io/gh/bjoluc/appcfg)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/appcfg.svg)](https://pypi.python.org/pypi/appcfg/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nFlexible hierarchic Python application configuration inspired by [node-config](https://github.com/lorenwest/node-config)\n\n## Motivation\n\nApplications (especially web services) often require certain configuration options to depend on the environment an application runs in (development, testing, production, etc.).\nFor instance, a database address config option may default to a local database server during development, a mock database server during testing, and yet another database server during production.\nIt may also need to be customizable via an environment variable.\n`appcfg` approaches scenarios like this and, similar to [node-config](https://github.com/lorenwest/node-config) for Node.js, allows to specify default configuration options for various environments and optionally override them by custom environment variables.\n\n## Getting Started\n\nLet\'s start by installing `appcfg` with ```pip install appcfg["yaml"]```, or simply ```pip install appcfg``` if you want to use the JSON format instead of YAML for config files.\n\nIn the top-level directory of your application (where the topmost `__init__.py` file is located), create a `config` directory.\nIf your application consists of a single Python file, just locate the `config` directory next to it.\nHere\'s an example project tree that we will refer to in the rest of this section:\n\n```diff\n ├── myproject\n │   ├── __init__.py\n+│   ├── config\n+|   |   └── ...\n |   ├── anothermodule.py\n │   └── myproject.py\n ├── tests\n │   ├── __init__.py\n │   └── test_myproject.py\n```\n\nWithin the `config` directory, create a `default.yml` file (or `default.json`, if you prefer that).\nThis file will hold your default configuration.\nIn the database example from the previous section, `default.yml` might look like this:\n\n```yaml\ndatabases:\n    mongo:\n        host: localhost\n        port: 27017\n        user: myuser\n        pass: mypassword\nsecrets:\n    mysecret: secret\n```\n\nWithin any module in the `myproject` package, we can now simply access that configuration as a `dict` object:\n\n```python\nfrom appcfg import get_config\n\nconfig = get_config(__name__)\nprint(config["databases"]["mongo"]["host"])\n```\n\n`__name__` is passed to `get_config` so that it can infer the project root path where the `config` directory is located.\nThis way, `appcfg` can be used independently in multiple projects loaded at the same time, and projects can also retrieve one another\'s configuration.\nFor instance, in `test_myproject.py`, the configuration of `myproject` could be retrieved with `get_config("myproject")`.\n\n## Environments\n\nLet\'s add an override config file for production, `production.yml`:\n\n```yaml\ndatabases:\n    mongo:\n        host: mongodb\n```\n\nAnd one for testing, `test.yml`:\n\n```yaml\ndatabases:\n    mongo:\n        host: mock\nsecret:\n    mysecret: wellknown\n```\n\nBy default, none of those config files will be used.\nHowever, an environment can be specified by setting the `ENV` environment variable (alternatively, `PY_ENV` or `ENVIRONMENT` are also supported).\nIn this case, the configuration options from the corresponding config file will be merged into the ones from `default.yml`.\nIf, for instance, `ENV` is set to `production`, the config dict returned from `get_config()` will contain all the values from `default.yml`, but `config["databases"]["mongo"]["host"]` will be set to `mongodb` instead of `localhost`.\nSimilarly, with `ENV=test`, `config["databases"]["mongo"]["host"]` would be `mock` and, `config["secret"]["mysecret"]` would be `wellknown`.\nIn case `ENV` is set but no corresponding config file is found, `get_config()` returns the options from the `default` config file.\n\n## Custom Environment Variables\n\nLet\'s say, we want the database host, user, and password, and the secret to be customizable via additional environment variables.\nThis can be achieved by adding an `env-vars.yml` config file with the following contents:\n\n```yaml\ndatabases:\n    mongo:\n        host: MONGO_HOST\n        user: MONGO_USER\n        pass: MONGO_PASS\nsecrets:\n    mysecret: MY_SECRET\n```\n\nThis way, if one of the specified environment variables is set, it will override the corresponding field\'s value from any other configuration file.\nOtherwise, that value is left untouched.\nFor instance, setting `MONGO_HOST=myhost` would result in `config["databases"]["mongo"]["host"]` to be `myhost`, ignoring `localhost` from `default.yml` or `mongodb` from `production.yml`.\n\n## Tips and Tricks\n\n### Environment Specification with pytest\n\nYou may wish to set `ENV=test` during unit tests without manually specifying it for every `pytest` invocation.\n[pytest-env](https://github.com/MobileDynasty/pytest-env) can do the job for you if you specify\n\n```ini\n[pytest]\nenv =\n    ENV=test\n```\n\nin your pytest configuration.\n\n<!-- BEING API DOC -->\n\n## API\n\n\n### get_config\n\nReturns a dict that contains the content of `default.json` or `default.yml` in the `config` directory within the root module\'s directory, inferred from `module_name`.\n\n\nIf an `ENV`, `PY_ENV`, or `ENVIRONMENT` variable is set (listed in the order of\nprecedence), and a config file with a base name corresponding to that variable\'s\nvalue is found, the contents of that config file are merged into the default\nconfiguration. Additionally, the environment variables specified in `env-vars.json`\nor `env-vars.yml` override any other configuration when they are set.\n\nIf none of `ENV`, `PY_ENV`, or `ENVIRONMENT` is set, only the `default` config file\nwill be loaded and optionally be overridden by custom environment variables as\nspecified in the `env-vars` config file. The `ENV` values "dev" and "develop" map to\nthe `development.json` or `development.yml` config file.\n\n**Arguments**:\n\n- `module_name` (`str`): The name of the module (or any of its submodules) for which\n  the configuration should be loaded for. `__name__` can be used when `get_config()`\n  is called from the Python project that contains the `config` directory. Note that\n  `appcfg` requires the `config` directory to be a direct child of the top-level\n  package directory, or, if not a package, the directory that contains the specified\n  module.\n\n- `cached` (`bool`): If `True` (the default), the configuration for each `config`\n  directory will only be loaded once and the same dict object will be returned by\n  every subsequent call to `get_config()`.\n\n<!-- END API DOC -->\n',
    'author': 'bjoluc',
    'author_email': 'mail@bjoluc.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bjoluc/appcfg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
